package droidface

import scala.collection.JavaConverters._
import scala.util.matching.Regex

import soot.{Unit => SootUnit, util => _, _}
import soot.toolkits.graph._
import soot.jimple._
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._
import soot.util.{JasminOutputStream}
import soot.options.Options

import java.io.{PrintWriter, OutputStreamWriter, FileOutputStream}

class GlobalTransform extends SceneTransformer {

  var facetedClasses: Map[Type, FacetedClass] = Map()
  var facetedFields: Map[SootField, FacetedField] = Map()
  var facetedMethods: Map[SootMethod, SootMethod] = Map()


  protected override def internalTransform(phase: String,
    options: java.util.Map[String, String]) = {
    Scene.v.loadClassAndSupport("java.lang.StringBuffer")
    Scene.v.getApplicationClasses.asScala filter isApplicationClass foreach
      facetizeFields
    Scene.v.getApplicationClasses.asScala filter isApplicationClass foreach
      facetizeMethods
    Util.debug(s"Finished initial transformation pass")
  }

  /* Generate a class to hold a faceted value of given (primitive) type.
   * This will generate a class that has one public field for each level
   * containing the value at that level.
   * This method also writes the class to the corresponding class file.
   */
  def genFacetedClass(t: Type): FacetedClass = {
    Scene.v().loadClassAndSupport("java.lang.Object");
    val nameTemplate = "Faceted" + t.toString.replaceAll("java.lang.", "").
      replaceAll("\\[\\]", "Arr")
    val className = freshClassName(nameTemplate)
    val newSootClass = new SootClass(className)
    val newClass = new FacetedClass(t, newSootClass)
    facetedClasses += ((t, newClass))
    // Scene.v.addClass(newSootClass)
    Scene.v.loadNecessaryClasses()
    newClass
  }

  def facetizeMethods(cls: SootClass): Unit = {
    cls.getMethods().asScala foreach { meth =>
      val newMethNeeded = newMethodNeeded(meth) &&
        !TaintTransformer.v.isEntrypoint(meth)
      if (newMethNeeded) {
        if (newMethNeeded) {
          createFacetedMethod(cls, meth)
        } else {
          Util.debug(s"No faceted copy needed for $meth")
        }
      }
    }
  }

  def createFacetedMethod(cls: SootClass, meth: SootMethod): Unit = {
    Util.debug(s"Creating faceted method for $meth")
    val (argTypes, retType) = facetedMethodType(meth)
    val newMeth = new SootMethod(meth.getName, argTypes.asJava, retType,
      meth.getModifiers, meth.getExceptions)
    facetedMethods += ((meth, newMeth))
    cls.addMethod(newMeth)

    if (meth.isConcrete) {
      val husk = TaintTransformer.v.createHusk(newMeth)
      newMeth.setActiveBody(husk)
    }
  }

  def facetizeFields(cls: SootClass) = {
    // get fields once to avoid ConcurrentModificationException:
    Util.debug(s"Adding faceted copies in class $cls")
    val flds = cls.getFields.asScala.toList filter
      (f => GlobalTransform.isInputType(f.getType))
    flds foreach { fld =>
      Util.debug(s"Adding faceted copies of field ${cls}.${fld}")
      val facetedFld = new FacetedField(fld, cls)
      facetedFields += ((fld, facetedFld))
    }
  }

  def freshClassName(pref: String): String = {
    var cnt = -1
    var freshName = pref
    val classes = Scene.v.getClasses.asScala
    while (classes.find(_.getName() == freshName) != None) {
      cnt += 1
      freshName = pref + "_" + cnt
    }
    freshName
  }

  def newMethodNeeded(meth: SootMethod): Boolean = {
    // Some exceptions when only partially transforming an app
    // Used to get CaffeineMark to run.
    val dontFacetExceptions: Seq[Regex] = Seq("testName".r, "execute".r,
      "cleanUp".r, "defaultMagnification".r, "initialize".r)
    val ignore: Boolean =
      Main.options.target == "com.android.cm3-1.apk" &&
      meth.getDeclaringClass.implementsInterface("com.android.cm3.BenchmarkAtom") &&
      dontFacetExceptions.exists(pat => pat.findFirstIn(meth.getName) != None)
    if (ignore) {
      Util.debug(s"Ignoring benchmark entry point $meth")
    }
    (GlobalTransform.isInputType(meth.getReturnType) ||
      (meth.getParameterTypes.asScala exists (GlobalTransform.isInputType(_)))) &&
    !ignore
  }

  def facetedMethodType(meth: SootMethod): (Seq[Type], Type) = {
    (meth.getParameterTypes.asScala map facetedType,
      facetedType(meth.getReturnType))
  }

  def facetedType(t: Type): Type =
    if (GlobalTransform.isInputType(t))
      GlobalTransform.v.getFacetedClass(t).getSootClass.getType
    else
      t


  def getFacetedClass(t: Type): FacetedClass = {
    if (!GlobalTransform.isInputType(t)) {
      Util.die(s"BUG: getFacetedClass called with non-input type $t")
    }
    facetedClasses.get(t) match {
      case Some(fclass) => fclass
      case None =>
        genFacetedClass(t)
    }
  }
  def isApplicationClass(cl: SootClass): Boolean = {
    !cl.toString.startsWith("android.support") &&
    !Main.options.ignoreClasses.contains(cl.getName)
  }

  def isApplicationMethod(meth: SootMethod): Boolean = {
    isApplicationClass(meth.getDeclaringClass) &&
      (Scene.v.getApplicationClasses.asScala exists (_.getMethods.asScala contains meth))
  }
}

object GlobalTransform {
  private val instance: GlobalTransform = new GlobalTransform()
  def v: GlobalTransform = instance

  def inputTypes: Seq[Type] = Seq(IntType.v, BooleanType.v,
    ShortType.v, LongType.v, FloatType.v, ByteType.v,
    DoubleType.v, RefType.v("java.lang.String"), RefType.v("java.lang.StringBuffer"),
    RefType.v("java.lang.Float"), RefType.v("java.lang.Double"), RefType.v("java.lang.Integer"),
    CharType.v)

  /* Check whether type is a possible input and hence needs to be faceted. */
  def isInputType(t: Type): Boolean =
    t match {
      case at: ArrayType => isInputType(at.getElementType)
      case t if inputTypes.contains(t) => true
      case _ => false
    }
}
