package droidface

import scala.annotation.tailrec
import scala.collection.JavaConverters._
import scala.collection.mutable.{Buffer, Queue, Map}
import scala.collection.Set
import scala.collection.mutable.HashSet
import scala.util.Random;
import util.control.Breaks._

import soot.{Unit => SootUnit, util => _, _}
import soot.toolkits.graph._
import soot.jimple._
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._

import Lattice.Level.{Value => _, _}

class LogTransformer extends BodyTransformer {

  protected override def internalTransform(body: Body, phase: String,
    options: java.util.Map[String, String]): Unit = {

    if (!GlobalTransform.v.isApplicationMethod(body.getMethod)) {
      return
    }
    val units: Queue[SootUnit] = Queue()
    var seen: Set[SootUnit] = Set()
    val cfg: BriefUnitGraph = new BriefUnitGraph(body)
    Util.overview.logMethod(body.getMethod, "original", body)
    // Util.debug("Iterating over units:")
    units.enqueue(cfg.getHeads.asScala: _*)
    while (!units.isEmpty) {
      breakable {
        var curUnit: SootUnit = units.dequeue()
        if (seen contains curUnit) {
          Util.debug(s"Already seen: $curUnit")
          break
        }
        seen += curUnit
        curUnit match {
          case ie: InvokeStmt =>
            val invokeExpr = ie.getInvokeExpr
            if (invokeExpr.getMethod.toString.contains("sendTextMessage")) {
              addSinks(body, ie)
            }
          case _ =>
        }

        cfg.getSuccsOf(curUnit).asScala foreach { u =>
          if (! (seen contains u))
            units.enqueue(u)
        }
      }
    }
    Util.overview.logMethod(body.getMethod, "transformed", body)
  }

  def addSinks(body: Body, ie: InvokeStmt) = {
    Util.debug(s"Adding log call for $ie")
    val invokeExpr = ie.getInvokeExpr
    val stringType: Type = RefType.v("java.lang.String")
    val logClass = Scene.v.loadClassAndSupport("android.util.Log")
    val logMeth = logClass.getMethod("d",
      Seq(stringType, stringType).asJava)
    val logCall = Jimple.v.newStaticInvokeExpr(logMeth.makeRef,
      Seq(StringConstant.v("DroidFace"),
        invokeExpr.getArg(2)).asJava)
    val logStmt = Jimple.v.newInvokeStmt(logCall)
    body.getUnits.insertBefore(logStmt, ie)

    val utilClass: SootClass = Scene.v.getClasses().asScala find (c =>
      c.getName() == "com.example.daniel.tainted.TaintDroidSinkUtil") get
    val outMeth = utilClass.getMethod("out",
      Seq(RefType.v("java.lang.String").asInstanceOf[Type]).asJava)
    body.getUnits.insertBefore(
      Jimple.v.newInvokeStmt(
        Jimple.v.newStaticInvokeExpr(outMeth.makeRef, invokeExpr.getArg(2))),
      ie)
  }

  def buildSendCall(body: Body, ie: InvokeStmt) = {
    val invokeExpr = ie.getInvokeExpr
    // add network send, so taintdroid recognizes this:
    def addUnit(u: SootUnit) = body.getUnits.insertBefore(u, ie)

    val sockClass = Scene.v.loadClassAndSupport("java.net.Socket")
    val printWriterClass = Scene.v.loadClassAndSupport("java.io.PrintWriter")
    val ioExClass = Scene.v.loadClassAndSupport("java.io.IOException")
    val osClass = Scene.v.loadClassAndSupport("java.io.OutputStream")
    // add socket local
    val lg = new soot.javaToJimple.LocalGenerator(body)
    val sockLocal = lg.generateLocal(sockClass.getType)
    val sockAssign = Jimple.v.newAssignStmt(sockLocal, Jimple.v.newNewExpr(sockClass.getType))
    addUnit(sockAssign)
    val sockInit = sockClass.getMethod("<init>", 
      Seq(RefType.v("java.lang.String"), IntType.v).asJava)
    addUnit(Jimple.v.newInvokeStmt(Jimple.v.newSpecialInvokeExpr(sockLocal,
      sockInit.makeRef, StringConstant.v("10.0.2.2"), IntConstant.v(8000))))
    
    val owLocal = lg.generateLocal(osClass.getType)
    val getOutputStreamMeth = sockClass.getMethod("getOutputStream", Seq().asJava)
    addUnit(Jimple.v.newAssignStmt(owLocal, 
      Jimple.v.newVirtualInvokeExpr(sockLocal, getOutputStreamMeth.makeRef)))
    val pwLocal = lg.generateLocal(printWriterClass.getType)
    addUnit(Jimple.v.newAssignStmt(pwLocal,
      Jimple.v.newNewExpr(printWriterClass.getType)))
    val pwInit = printWriterClass.getMethod("<init>", Seq(osClass.getType.asInstanceOf[Type]).asJava)
    addUnit(Jimple.v.newInvokeStmt(
      Jimple.v.newSpecialInvokeExpr(pwLocal, pwInit.makeRef, owLocal)))

    // val add write call
    val writeMeth = printWriterClass.getMethod("write", Seq(RefType.v("java.lang.String").asInstanceOf[Type]).asJava)
    val flushMeth = printWriterClass.getMethod("flush", Seq().asJava)
    val pwCloseMeth = printWriterClass.getMethod("close", Seq().asJava)
    val sockCloseMeth = sockClass.getMethod("close", Seq().asJava)
    addUnit(Jimple.v.newInvokeStmt(
      Jimple.v.newVirtualInvokeExpr(pwLocal, writeMeth.makeRef, invokeExpr.getArg(2))))
    addUnit(Jimple.v.newInvokeStmt(
      Jimple.v.newVirtualInvokeExpr(pwLocal, flushMeth.makeRef)))
    addUnit(Jimple.v.newInvokeStmt(
      Jimple.v.newVirtualInvokeExpr(pwLocal, pwCloseMeth.makeRef)))
    val closeStmt = Jimple.v.newInvokeStmt(
      Jimple.v.newVirtualInvokeExpr(sockLocal, sockCloseMeth.makeRef))
    addUnit(closeStmt)
    val landingNop = Jimple.v.newNopStmt
    addUnit(landingNop)
    // add exception handler
    val exceptionClass = Scene.v.loadClassAndSupport("java.io.IOException")
    val throwableClass = Scene.v.loadClassAndSupport("java.lang.Throwable")
    val stackTraceMeth = throwableClass.getMethod("printStackTrace", Seq().asJava)
    val excLocal = lg.generateLocal(exceptionClass.getType)
    val excAssign = Jimple.v.newIdentityStmt(excLocal, Jimple.v.newCaughtExceptionRef)
    val excHandlerStart: SootUnit =
      Jimple.v.newInvokeStmt(Jimple.v.newVirtualInvokeExpr(excLocal, stackTraceMeth.makeRef))
    val jmpToNop = Jimple.v.newGotoStmt(landingNop)
    body.getUnits.addLast(excAssign)
    body.getUnits.insertAfter(Seq(excHandlerStart, jmpToNop).asJava, excAssign)
    body.getTraps().addLast(Jimple.v.newTrap(exceptionClass, sockAssign, closeStmt, excAssign))
  }

}
