package droidface

import scala.annotation.tailrec
import scala.collection.JavaConverters._
import scala.collection.mutable.{Buffer, Queue, Map}
import scala.collection.Set
import scala.collection.mutable.HashSet
import scala.util.Random;
import util.control.Breaks._

import soot.{Unit => SootUnit, util => _, _}
import soot.toolkits.graph._
import soot.jimple._
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._

import Lattice.Level.{Value => _, _}

class TaintTransformer extends BodyTransformer {
  
  var cfg: UnitGraph = null
  var body: Body = null
  var varCopies : Map[Value, FacetedVar] = Map()
  var lg: soot.javaToJimple.LocalGenerator = null
  var oldMeth: SootMethod = null
  var firstNonId: SootUnit = null

  // this is an unclean way to handle this, but
  // refactoring this into something cleaner is
  // a lot of effort
  var exceptionCalls: Set[(SootUnit, SootClass)] = Set()

  var wrapperClasses: Set[SootClass] = null

  protected override def internalTransform(body: Body, phase: String,
    options: java.util.Map[String, String]): Unit = {
    // Needs to be initialized here, since soot is not initialized
    // sufficiently before.
    if (wrapperClasses == null) {
      wrapperClasses = Set("java.lang.Float", "java.lang.Integer",
        "java.lang.Double", "java.lang.Boolean") map (Scene.v.loadClassAndSupport(_))
    }
    this.body = body
    exceptionCalls = Set()
    val clasName = body.getMethod.getDeclaringClass.toString
    if (Main.options.ignoreClasses.contains(clasName)) {
      Util.indented { Util.debug(s"Ignoring class $clasName") }
      return
    }
    if (!GlobalTransform.v.isApplicationMethod(body.getMethod)) {
      Util.indented {
        // Util.debug(s"Skipping non-application method ${body.getMethod}\n")
      }
      return
    }
    if (GlobalTransform.v.facetedMethods.values.exists (_ == body.getMethod)) {
      Util.debug(s"Skipping newly created method ${body.getMethod}")
      return
    }
    if (isOutputMethod(body.getMethod) != None ||
      isInputMethod(body.getMethod) != None) {
      return
    }
    Util.debug(s"Analyzing method ${body.getMethod} " ++
      s"[${body.getMethod.getDeclaringClass()}]")
    lg = new soot.javaToJimple.LocalGenerator(body)
    Util.overview.logMethod(body.getMethod, "original", body)
    /* Create a copy of the method with faceted types, if needed: */
    oldMeth = body.getMethod
    val newMethNeeded = GlobalTransform.v.facetedMethods.get(oldMeth) match {
      case Some(newMeth) =>
        // body.getMethod.getDeclaringClass.addMethod(newMeth)
        newMeth.setActiveBody(body)
        true
      case None => false
    }
    Util.indented {
      handleBody(body)
      Util.overview.logMethod(oldMeth, "transformed", body)
      if (newMethNeeded) {
        val husk = createHusk(oldMeth)
        oldMeth.setActiveBody(husk)
      }
      Util.debug(s"Validating body for ${body.getMethod}")
      val errList: java.util.List[soot.validation.ValidationException] = new java.util.Vector()
      body.validate(errList)
      if (!errList.isEmpty) {
        errList.asScala foreach (e => Util.warn(s"Validation error ${e}"))
        Util.die(s"Validation failed for body of ${body.getMethod}")
      }
      Util.debug(s"Finished processing body for ${body.getMethod}\n")
    }
    Util.setIndent(0)
  }

  def isEntrypoint(meth: SootMethod): Boolean = {
    // quick-and-dirty heuristic for detecting android event
    // handlers
    val eventHandlers = Seq("onCreateOptionsMenu",
      "onOptionsItemSelected", "onCreate")
    meth.isEntryMethod || eventHandlers.exists (meth.toString.contains(_))
  }

  def createHusk(oldMeth: SootMethod): Body = {
    // Create a "husk" for the old method
    // TODO: figure out how to delete the old method without making
    // the rest of soot crash
    val husk = Jimple.v.newBody(oldMeth)
    val oldLg = new soot.javaToJimple.LocalGenerator(husk)
    // non-static methods need a this local
    val thisLoc: Option[Local] =
      if (!oldMeth.isStatic) {
        val classType = oldMeth.getDeclaringClass().getType
        val thisL = oldLg.generateLocal(classType)
        husk.getUnits.addLast(Jimple.v.newIdentityStmt(thisL,
          Jimple.v.newThisRef(classType)))
        Some(thisL)
      } else { None }
    (0 until oldMeth.getParameterCount()) foreach { i =>
      val t = oldMeth.getParameterType(i)
      val loc = oldLg.generateLocal(t)
      husk.getUnits.addLast(Jimple.v.newIdentityStmt(loc,
        Jimple.v.newParameterRef(t, i)))
    }
    // Call init on this, if we are in a constructor:
    thisLoc match {
      case Some(thisL) if oldMeth.isConstructor =>
        // We might not have a constructor that takes no argument,
        // so we call the one from Object
        
          val objectClass =
            if (oldMeth.getDeclaringClass.getSuperclass.getType ==
              RefType.v("java.lang.Thread")) {
              Scene.v.loadClassAndSupport("java.lang.Thread")
            } else {
              Scene.v.loadClassAndSupport("java.lang.Object")
            }
          val constr = objectClass.getMethod("<init>", Seq().asJava)
          husk.getUnits.addLast(Jimple.v.newInvokeStmt(
            Jimple.v.newSpecialInvokeExpr(thisL, constr.makeRef)))
      case _ =>
    }
    val retType = oldMeth.getReturnType
    val retVals: Map[Type, Value] =
      Map((IntType.v, IntConstant.v(0)),
        (ShortType.v, IntConstant.v(0)),
        (LongType.v, LongConstant.v(0)),
        (BooleanType.v, IntConstant.v(0)),
        (FloatType.v, FloatConstant.v(0.0f)),
        (DoubleType.v, DoubleConstant.v(0.0d)))
    // TODO: arrays
    husk.getUnits.addLast(
      if (retType == VoidType.v)
        Jimple.v.newReturnVoidStmt()
      else Jimple.v.newReturnStmt(retVals.getOrElse(retType, NullConstant.v)))
    husk
  }


  def adjustMethodType(meth: SootMethod): Unit = {
    Util.debug(s"Adjusting type of method $meth")
    // Util.debug(s"Body: ${meth.get}")
    val retType = meth.getReturnType
    if (GlobalTransform.isInputType(retType)) {
      val facetedRetType = GlobalTransform.v.getFacetedClass(retType)
      meth.setReturnType(facetedRetType.getSootClass.getType)
    }
    def handleParamType(t: Type) =
      if (GlobalTransform.isInputType(t))
        GlobalTransform.v.getFacetedClass(t).getSootClass.getType
      else
        t
    meth.setParameterTypes(meth.getParameterTypes().asScala map handleParamType asJava)
    body.getClass.getMethods()
    // cfg = new BriefUnitGraph(body)
  }

  def isFacetedLocal(v: Value): Option[FacetedVar] = varCopies.get(v)

  def handleBody(body : Body): Unit = {

    /* Make copies of locals */
    body.getLocals().asScala filter
      (l => GlobalTransform.isInputType(l.getType)) foreach { loc =>
        Util.debug(s"Adding copy of local $loc [${loc.getType}]:")
        val newLoc = new FacetedLocal(loc, body)
        Util.indented {
          Lattice.levels foreach { l =>
            Util.debug(s"Copy at level $l: ${newLoc.levelVal(l)}")
          }
        }
        varCopies += ((loc, newLoc))
      }

    // recomputeCFG()
    // initFVPrefix()
    val units: Queue[SootUnit] = Queue()
    var seen: Set[SootUnit] = Set()
    // cfg = new BriefUnitGraph(body)
    cfg = new ExceptionalUnitGraph(body)
    firstNonId = findFirstNonId(cfg.getHeads.asScala)
    Util.debug(s"First non-identity statement: $firstNonId")
    Util.debug("Iterating over units:")
    units.enqueue(cfg.getHeads.asScala: _*)
    while (!units.isEmpty) {
      breakable {
        var curUnit: SootUnit = units.dequeue()
        if (seen contains curUnit) {
          Util.debug(s"Already seen: $curUnit")
          break
        }
        Util.debug(s"Analyzing unit: $curUnit [${curUnit.getClass}]")
        seen += curUnit
        Util.incIndent()
        handleUnit(curUnit)
        Util.decIndent()
        cfg.getSuccsOf(curUnit).asScala foreach { u =>
          if (! (seen contains u))
            units.enqueue(u)
        }
      }
    }
    // Add try-catch blocks:
    exceptionCalls.foreach (_ match { case (u, exc) =>
      addTryCatch(u, exc)
    })
  }

  def findFirstNonId(starts: Seq[SootUnit]): SootUnit = {
    if (starts.length != 1) {
      Util.die(s"Method ${body.getMethod} has multiple entry points: $starts")
    } else {
      val u = starts.head
      u match {
        case id: IdentityStmt => 
          val next = cfg.getSuccsOf(u).asScala
          if (next.length == 0) {
            u
          } else {
            findFirstNonId(next)
          }
        case _ => u
      }
    }
  }

  def handleUnit(unit: SootUnit) : Unit = {
    val clasPattern = """class soot\.jimple\.internal\.(.*)$""".r
    val clas = clasPattern replaceAllIn
      (unit.getClass.toString, _ match { case clasPattern(x) => x })
    val lg = new soot.javaToJimple.LocalGenerator(body)
    unit match {
      case assign: AssignStmt => handleAssign(assign)
      case ifStmt: IfStmt =>
        ifStmt.getCondition match {
          case invExpr: InvokeExpr =>
            // this is never produced by jimple, as far as I can tell
            Util.warn(s"InvokeExpr in guard of if statement: $ifStmt")
          case expr =>
            ifStmt.setCondition(substLevel(Lattice.top, expr))
        }
      case inv: InvokeStmt =>
        val (_, newStmts) = handleInvokeExpr(inv.getInvokeExpr(), None)
        body.getUnits.insertAfter(newStmts.asJava, inv)
        body.getUnits.remove(inv)
      case id: IdentityStmt =>
        id.getRightOp match {
          case param: ParameterRef =>
            if (GlobalTransform.isInputType(param.getType)) {
              if (!isEntrypoint(body.getMethod) &&
                GlobalTransform.v.facetedMethods.get(oldMeth) != None) {
                val facetedClass = GlobalTransform.v.getFacetedClass(param.getType)
                val floc = lg.generateLocal(facetedClass.getSootClass.getType)
                val fIdStmt = Jimple.v.newIdentityStmt(floc, param)
                body.getUnits.insertAfter(fIdStmt, id)
                body.getUnits.remove(id)
                isFacetedLocal(id.getLeftOp) match {
                  case Some(fv) =>
                    // val nonIdStmt = 
                    Lattice.levels foreach { l =>
                      body.getUnits.insertBefore(
                        Jimple.v.newAssignStmt(fv.levelVal(l),
                          Jimple.v.newInstanceFieldRef(floc,
                            facetedClass.getLevelField(l))),
                        firstNonId)
                    }
                  case None => Util.die(s"Found non-faceted value in IdentityStmt: $id")
                }
              } else {
                // We got a primitive type as an argument, but not in
                // a method that has been faceted; copy it to all
                // locals; we need to use dummy values if entry points
                // somehow get secrets as arguments
                isFacetedLocal(id.getLeftOp) match {
                  case Some(fv) =>
                    Lattice.levels foreach { l =>
                      body.getUnits.insertAfter(
                        Jimple.v.newAssignStmt(fv.levelVal(l),
                          id.getLeftOp), id)
                    }
                  case None => Util.die(s"Found non-faceted value in IdentityStmt: $id")
                }
              }
            }
          case _ =>
        }
      case _: ReturnVoidStmt =>
      case exitStmt: ReturnStmt =>
        if (GlobalTransform.isInputType(exitStmt.getOp.getType)) {
          val inFacetedMeth = GlobalTransform.v.facetedMethods.get(oldMeth) != None
          exitStmt.getOp match {
            case loc: Local => 
              isFacetedLocal(exitStmt.getOp) match {
                case Some(fv) =>
                  // If we changed the method type to return something faceted,
                  // we need to put the values in a struct
                  if (inFacetedMeth) {
                    val (loc, initStmts) = createFacetedStruct(fv)
                    body.getUnits.insertBefore(initStmts.asJava, exitStmt)
                    exitStmt.setOp(loc)
                  } else {
                    // the return type is an input type, but we are in an entry method
                    // default to returning the high or low part
                    exitStmt.setOp(
                      fv.levelVal(if (Main.options.defaultLow)
                        Lattice.bot
                      else Lattice.top))
                  }
                case None =>
                  Util.die(s"Argument to return statement not faceted.")
              }
            case cnst: Constant  =>
              if (inFacetedMeth) {
                val (loc, initStmts) = createFacetedStruct(new FacetedConst(cnst,
                  oldMeth.getReturnType))
                body.getUnits.insertBefore(initStmts.asJava, exitStmt)
                exitStmt.setOp(loc)
              } else {
                // Input type as return, but we didn't change the return type
                // just leave the constant as is
              }
            case e => Util.warn(s"Unhandled argument to return: $e [${e.getClass}]")
          }
        }
      case _: GotoStmt =>
      case s => Util.warn(s"Unhandled $clas: $unit")
    }
  }


  def handleAssign(assign: AssignStmt): Unit = {
    assign.getRightOp match {
      /* Since we must not duplicate method calls, we treat
       * invoke expressions differently:
       * - Use handleInvokeExpr to modify arguments
       * - If the method returns an input type, represented as a faceted
       *   class, then split the result into our local copies.
       * Note that this relies on Jimple producing fairly simple right-hand sides
       * of assignments; they contain either a method call at the outermost
       * level or no method call at all.
       */
      case ie: InvokeExpr =>
        val resultType = ie.getType
        val (optLoc, newStmts) = handleInvokeExpr(ie, assign.getLeftOp match {
          case l: Local => Some(l)
            // note that assigning the result of an invoke expression to a non-local
            // should be ruled out by jimple
          case _ => None
        })
        body.getUnits.insertAfter(newStmts.asJava, assign)
        body.getUnits.remove(assign)
        isFacetedLocal(assign.getLeftOp) match {
          case Some(fv) =>
            optLoc match {
              case Some(resultLoc) =>
                // In the case of built-in methods, we may not get a faceted
                // value back:
                if (GlobalTransform.isInputType(resultLoc.getType)) {
                  // In this case we just copy the result to all the levels, to
                  // avoid crashes:
                  Lattice.levels foreach { l =>
                    body.getUnits.insertAfter(
                      Jimple.v.newAssignStmt(fv.levelVal(l), resultLoc),
                      newStmts.last)
                  }
                } else {
                  // Otherwise, we got a faceted structure back, so we
                  // just copy the respective fields to the locals
                  val valClass = GlobalTransform.v.getFacetedClass(resultType)
                  Lattice.levels foreach { l =>
                    val assignResult =
                      Jimple.v.newAssignStmt(fv.levelVal(l),
                        Jimple.v.newInstanceFieldRef(resultLoc,
                          valClass.getLevelField(l))).asInstanceOf[SootUnit]
                    body.getUnits.insertAfter(assignResult, newStmts.last)
                  }
                }
              case _ => // this means handleInvokeExpr already took care of
                // the assignment
            }
          case None =>
            optLoc match {
              case Some(loc) =>
                // TODO: this generates one local too many in the case of non-input
                // types; maybe optimize if there is time.
                body.getUnits.insertAfter(
                  Jimple.v.newAssignStmt(assign.getLeftOp, loc), newStmts.last)
              case None =>
                  Util.die("No local for faceted result type of method ${ie.getMethod}")
            }
        }
      /* Otherwise, just duplicate the expression at each level: */
      case _ =>
        isFacetedLocal(assign.getLeftOp) match {
          case Some(fv) =>
            Lattice.levels foreach { l =>
              val lhs = fv.levelVal(l)
              val rhs = substLevel(l, assign.getRightOp)
              val newStmt = Jimple.v.newAssignStmt(lhs, rhs)
              body.getUnits.insertAfter(newStmt, assign)
            }
            body.getUnits.remove(assign)
          case None =>
            if (GlobalTransform.isInputType(assign.getLeftOp.getType)) {
                Lattice.levels foreach { l =>
                  val lhs = substLevel(l, assign.getLeftOp)
                  val rhs = substLevel(l, assign.getRightOp)
                  val newStmt = Jimple.v.newAssignStmt(lhs, rhs)
                  body.getUnits.insertAfter(newStmt, assign)
                }
                body.getUnits.remove(assign)
            } else {
              Util.debug(s"Non-invoke assign of non-faceted type: $assign [${assign.getLeftOp.getType}]")
              val fallbackLevel = if (Main.options.defaultLow) Lattice.bot else Lattice.top
              val newStmt = Jimple.v.newAssignStmt(substLevel(fallbackLevel, assign.getLeftOp),
                substLevel(fallbackLevel, assign.getRightOp))
              body.getUnits.insertAfter(newStmt, assign)
              body.getUnits.remove(assign)
            }
        }
    }
  }

  /* Transform an invoke expression and return a local containing the
     result, if any, along with a sequence of statements to be inserted
     in its place.
   - If the expression is an input expression, perform the input,
     then return a faceted value (as a FacetedClass).
   - If the expression is an output, then retrieve the argument values at
     the corresponding level.
   - If the expression is neither, then pass faceted values as arguments
     and return a faceted result, if the called method is defined by
     the application. Otherwise, just pass in the top values as an unsound
     fallback.
   */
  def handleInvokeExpr(invokeExpr: InvokeExpr, originalLoc: Option[Local]):
      (Option[Local], Seq[SootUnit]) = {
    val res = handleInvokeAux(invokeExpr, originalLoc)
    // If we created a new method, then point this invokeexpr to the new one
    GlobalTransform.v.facetedMethods.get(invokeExpr.getMethod) match {
      case Some(fmeth) if (isInput(invokeExpr) == None) &&
          (isOutput(invokeExpr) == None) =>
        Util.debug(s"Changing invokeExpr $invokeExpr to use faceted method $fmeth")
        invokeExpr.setMethodRef(fmeth.makeRef)
      case _ =>
    }
    res
  }

  def handleInvokeAux(invokeExpr: InvokeExpr, originalLoc: Option[Local]):
      (Option[Local], Seq[SootUnit]) = {
    val lg = new soot.javaToJimple.LocalGenerator(body)
    val resultType = invokeExpr.getType
    // assumption: no method call is both a source and a sink:
    (isInput(invokeExpr), isOutput(invokeExpr)) match {
      case (Some(inLevel), _) => {
        Util.debug(s"Found input expression $invokeExpr of level $inLevel")
        if (invokeExpr.getArgCount != 0) {
          Util.die("Don't know how to handle input methods with args")
        } else {
          val facetedClass = GlobalTransform.v.getFacetedClass(resultType)
          val facetedSootClass = facetedClass.getSootClass
          Util.debug(s"Using $facetedClass with fields ${facetedSootClass.getFields}")
          val fresultLoc = lg.generateLocal(facetedSootClass.getType)
          val fresultNewStmt = Jimple.v.newAssignStmt(fresultLoc,
            Jimple.v.newNewExpr(facetedSootClass.getType))
          val fresultInitStmt = Jimple.v.newInvokeStmt(
            Jimple.v.newSpecialInvokeExpr(fresultLoc,
              facetedClass.getConstructor.makeRef))
          val origResultLoc = lg.generateLocal(invokeExpr.getType())
          val resultAssign = Jimple.v.newAssignStmt(origResultLoc, invokeExpr)
          var levelAssigns: Buffer[SootUnit] = Buffer()
          Lattice.levels foreach { l =>
            val (dummyVal, stmts) = dummyValue(resultType)
            levelAssigns ++= stmts
            levelAssigns += Jimple.v.newAssignStmt(
              Jimple.v.newInstanceFieldRef(fresultLoc,
                facetedSootClass.getField(l.toString, resultType).makeRef),
              if (Lattice.le(inLevel, l)) origResultLoc
              else dummyVal)
          }
          // val finalLoc = lg.generateLocal(facetedClass.getType)
          (Some(fresultLoc),
            Seq(fresultNewStmt, fresultInitStmt, resultAssign) ++ levelAssigns.toSeq)
        }
      }
      case (_, Some(outLevel)) => {
        Util.debug(s"Found output expression $invokeExpr with level $outLevel")
        for (i <- 0 to invokeExpr.getArgCount() - 1) {
          invokeExpr.setArg(i, substLevel(outLevel, invokeExpr.getArg(i)))
        }
        // Add a log call if we are sending an SMS:
        val logCall = 
          if (invokeExpr.getMethod.toString.contains("sendTextMessage")) {
            val stringType: Type = RefType.v("java.lang.String")
            val logClass = Scene.v.loadClassAndSupport("android.util.Log")
            val logMeth = logClass.getMethod("d",
              Seq(stringType, stringType).asJava)
            val logCall = Jimple.v.newStaticInvokeExpr(logMeth.makeRef,
              Seq(StringConstant.v("DroidFace"),
                invokeExpr.getArg(2)).asJava)
            Seq(Jimple.v.newInvokeStmt(logCall))
          } else {
            Seq()
          }
        (None, logCall ++ Seq(Jimple.v.newInvokeStmt(invokeExpr)))
        // todo: result handling
      }
      case _ => {
        Util.debug(s"Handling non-IO invoke expression $invokeExpr [${invokeExpr.getMethod.getSignature}]")
        val methRef = invokeExpr.getMethodRef()
        if (GlobalTransform.v.isApplicationMethod(invokeExpr.getMethod)) {
          Util.debug(s"Call to application method: ${invokeExpr.getMethod}")
          // pass in all primitive arguments as faceted values as a FacetedClass instance
          val initStmts = (0 to invokeExpr.getArgs.asScala.size - 1) flatMap {i =>
            val arg = invokeExpr.getArg(i)
            if (GlobalTransform.isInputType(arg.getType)) {
              isFacetedLocal(arg) match {
                case Some(fv) =>
                  val (floc, initStmts) = createFacetedStruct(fv)
                  Util.debug(s"Setting argument $i of $invokeExpr to $floc")
                  invokeExpr.setArg(i, floc)
                  initStmts
                case None => arg match {
                  case const: Constant =>
                    val (floc, initStmts) = createFacetedStruct(new FacetedConst(const))
                    Util.debug(s"Setting argument $i of $invokeExpr to $floc")
                    invokeExpr.setArg(i, floc)
                    initStmts
                  case _ => Util.die(s"Unhandled argument type: $arg")
                }
              }
            } else {
              // some non-input type, don't change it:
              Seq()
            }
          }
          if (invokeExpr.getType != VoidType.v) {
            val locType = GlobalTransform.v.facetedMethods.get(invokeExpr.getMethod) match {
              case Some(fmeth) => fmeth.getReturnType
              case None => invokeExpr.getType
            }
            val loc = lg.generateLocal(locType)
            // TODO: this creates some unnecessary local and assignment
            // statement and should be optimized at some point:
            val assignLoc = Jimple.v.newAssignStmt(loc, invokeExpr)
            (Some(loc), initStmts ++ Seq(assignLoc))
          } else {
            (None, initStmts ++ Seq(Jimple.v.newInvokeStmt(invokeExpr)))
          }
        } else {
          // Special-case some known methods:
          Util.debug(s"Handling non-application method call ${invokeExpr.getMethod.toString}")
          Scene.v.loadClassAndSupport("java.lang.System")

          val stringClass = Scene.v.loadClassAndSupport("java.lang.String")
          val stringLike = Set(stringClass, Scene.v.loadClassAndSupport("java.lang.StringBuffer"))

          invokeExpr.getMethod match {
            case m if m == Scene.v.getMethod("<java.lang.System: void arraycopy(java.lang.Object,int,java.lang.Object,int,int)>") =>
              return duplicateCall(invokeExpr, originalLoc)
            case m if m.toString.contains("java.util.Arrays") &&
                m.toString.contains("toString") =>
              Util.debug(s"Duplicating call to Arrays.toString: $m")
              return duplicateCall(invokeExpr, originalLoc)
            case m if stringLike.exists(m.getDeclaringClass.equals(_)) =>
              return duplicateCall(invokeExpr, originalLoc)
              /*
              val patterns: Seq[String] = Seq("toCharArray", "<init>", "append", "indexOf", "length", "toString",
                "substring")
              if (patterns.exists(m.toString.contains(_))) {
                Util.debug(s"Duplicating string-related call $invokeExpr")
                return duplicateCall(invokeExpr, originalLoc)
              }
               */
            case m if wrapperClasses.exists(m.getDeclaringClass.equals(_)) =>
                return duplicateCall(invokeExpr, originalLoc)
            case _ =>
          }
          val fallbackLevel = if (Main.options.defaultLow) Lattice.bot else Lattice.top
          if (!Main.options.defaultLow) Util.unsound()
          val args: Seq[Value] =
              invokeExpr.getArgs.asScala map (substLevel(fallbackLevel, _))
          val invExpr = invokeExpr match {
            case staticInv: StaticInvokeExpr =>
              Jimple.v.newStaticInvokeExpr(methRef, args.asJava)
            case di: DynamicInvokeExpr =>
              Jimple.v.newDynamicInvokeExpr(methRef, args.asJava,
                di.getBootstrapMethodRef(), di.getBootstrapArgs())
              // If the method that is called depends on a secret, then this is an implict flow
              // so we can use the high copy here
            case si: SpecialInvokeExpr =>
              Jimple.v.newSpecialInvokeExpr(substLevel(Lattice.top, si.getBase).asInstanceOf[Local],
                methRef, args.asJava)
            case vi: VirtualInvokeExpr =>
              Jimple.v.newVirtualInvokeExpr(substLevel(Lattice.top, vi.getBase).asInstanceOf[Local],
                methRef, args.asJava)
            case ii: InterfaceInvokeExpr =>
              Jimple.v.newInterfaceInvokeExpr(substLevel(Lattice.top, ii.getBase).asInstanceOf[Local],
                methRef, args.asJava)
            case e =>
              Util.die(s"Unhandled type of invoke expression $e [${e.getClass}]")
          }
          if (invokeExpr.getType != VoidType.v) {
            val resultLoc = lg.generateLocal(invokeExpr.getType)
            (Some(resultLoc), Seq(Jimple.v.newAssignStmt(resultLoc, invExpr)))
          } else {
            (None, Seq(Jimple.v.newInvokeStmt(invExpr)))
          }
        }
      }
    }
  }

  def duplicateCall(ie: InvokeExpr, originalLoc: Option[Local]): (Option[Local], Seq[SootUnit]) = {
    var stmts: Buffer[SootUnit] = Buffer()
    val methRef = ie.getMethodRef
    Util.debug(s"Duplicating call: $ie")
    Lattice.levels foreach { l =>
      // TODO: avoid duplication
      val args: Seq[Value] = ie.getArgs.asScala map (substLevel(l, _))
      val invExpr: InvokeExpr = ie match {
        case staticInv: StaticInvokeExpr =>
          Jimple.v.newStaticInvokeExpr(methRef, args.asJava)
        case di: DynamicInvokeExpr =>
          Jimple.v.newDynamicInvokeExpr(methRef, args.asJava,
            di.getBootstrapMethodRef(), di.getBootstrapArgs())
        case si: SpecialInvokeExpr =>
          Jimple.v.newSpecialInvokeExpr(substLevel(l, si.getBase).asInstanceOf[Local],
            methRef, args.asJava)
        case vi: VirtualInvokeExpr =>
          Jimple.v.newVirtualInvokeExpr(substLevel(l, vi.getBase).asInstanceOf[Local],
            methRef, args.asJava)
        case ii: InterfaceInvokeExpr =>
          Jimple.v.newInterfaceInvokeExpr(substLevel(l, ii.getBase).asInstanceOf[Local],
            methRef, args.asJava)
        case e =>
          Util.die(s"Unhandled type of invoke expression $e [${e.getClass}]")
      }
      val stmt =
        ie.getType match {
          case t if t == VoidType.v => Jimple.v.newInvokeStmt(invExpr)
          case t if GlobalTransform.isInputType(ie.getType) =>
            originalLoc match {
              case Some(oLoc) =>
                isFacetedLocal(originalLoc.get) match {
                  case Some(fv) => Jimple.v.newAssignStmt(fv.levelVal(l), invExpr)
                  case None => Util.die(s"Result local for $ie has no faceted copies")
                }
              case None =>
                Util.warn(s"No result local for invoke expression $ie")
                Jimple.v.newInvokeStmt(invExpr)
            }
          case _ => Util.die(s"Unhandled return type of built-in method: ${ie.getMethod}")
        }
      if (l == Lattice.bot) {
        ie.getMethod match {
          case m if wrapperClasses.exists(m.getDeclaringClass.equals(_))
              && m.toString.contains("valueOf") =>
            addExceptionHandling(stmt,
              Scene.v.loadClassAndSupport("java.lang.NumberFormatException"))
          case _ =>
        }   
      }
      stmts += stmt
    }
    (None, stmts)
  }

  def addExceptionHandling(u: SootUnit, exc: SootClass) =
    exceptionCalls += ((u, exc))

  def substLevel(l: Level, rhs: Value): Value = {
    Util.debug(s"Substituting for level $l in $rhs [${rhs.getClass()}]")
    rhs match {
      case const: Constant => const
      case loc: Local =>
        isFacetedLocal(loc) match {
          case Some(fv) => fv.levelVal(l)
          case None => loc
        }
      case bexpr: BinopExpr =>
        val constr: (Value, Value) => Value = bexpr match {
          case expr: AddExpr => Jimple.v.newAddExpr
          case expr: MulExpr => Jimple.v.newMulExpr
          case expr: DivExpr => Jimple.v.newDivExpr
          case expr: SubExpr => Jimple.v.newSubExpr
          case expr: LtExpr => Jimple.v.newLtExpr
          case expr: EqExpr => Jimple.v.newEqExpr
          case expr: LeExpr => Jimple.v.newLeExpr
          case expr: GtExpr => Jimple.v.newGtExpr
          case expr: GeExpr => Jimple.v.newGeExpr
          case expr: NeExpr => Jimple.v.newNeExpr
          case expr: CmpgExpr => Jimple.v.newCmpgExpr
          case expr: CmplExpr => Jimple.v.newCmplExpr
          case expr: CmpExpr => Jimple.v.newCmpExpr
          case expr: RemExpr => Jimple.v.newRemExpr
          case expr: ShrExpr => Jimple.v.newShrExpr
          case expr: ShlExpr => Jimple.v.newShlExpr
          case expr: XorExpr => Jimple.v.newXorExpr
          case expr: AndExpr => Jimple.v.newAndExpr
          case expr: OrExpr => Jimple.v.newOrExpr
        }
        constr(substLevel(l, bexpr.getOp1), substLevel(l, bexpr.getOp2))
      case uexpr: UnopExpr =>
        val constr: Value => Value = uexpr match {
          case expr: NegExpr => Jimple.v.newNegExpr
          case expr: LengthExpr => Jimple.v.newLengthExpr
        }
        constr(substLevel(l, uexpr.getOp))
      case fldRef: FieldRef =>
        GlobalTransform.v.facetedFields.get(fldRef.getField()) match {
          case Some(ffld) =>
            fldRef match {
              case i: InstanceFieldRef =>
                Jimple.v.newInstanceFieldRef(i.getBase, ffld.levelRef(l))
              case s: StaticFieldRef =>
                Util.debug(s"Found static field reference $s [${s.getClass}]")
                Jimple.v.newStaticFieldRef(ffld.levelRef(l))
              case _ =>
                Util.die(s"Unhandled field access: $fldRef [${fldRef.getClass}]")
            }
          case None =>
            fldRef
        }
      case newArr: NewArrayExpr =>
        Jimple.v.newNewArrayExpr(newArr.getBaseType,
          substLevel(l, newArr.getSize))
      case arrRef: ArrayRef =>
        if (!GlobalTransform.isInputType(arrRef.getType)) {
          return Jimple.v.newArrayRef(arrRef.getBase, substLevel(l, arrRef.getIndex))
        }
        arrRef.getBase match {
          case arrLoc: Local =>
            isFacetedLocal(arrLoc) match {
              case Some(floc) =>
                Jimple.v.newArrayRef(floc.levelVal(l), substLevel(l, arrRef.getIndex))
              case None => Util.die(s"No faceted local for array $arrLoc")
            }
          case expr => Util.die(s"Unhandled array ref: $expr [${expr.getClass}]")
        }
      case cast: CastExpr => 
        Jimple.v.newCastExpr(substLevel(l, cast.getOp), cast.getCastType)
      case _ => {
        Util.warn(s"substLevel: Unhandled expression type: $rhs [${rhs.getClass}]")
        rhs
      }
    }
  }

  def isInputMethod(m: SootMethod): Option[Level] = {
    val methName = m.toString
    val inputPatterns = Seq("getSecret", "getDeviceId",
      "getLatitude", "getLongitude", "getSimSerialNumber")
    if (inputPatterns.exists (methName.contains(_))) {
      Some(H)
    } else {
      None
    }
  }

  def addTryCatch(u: SootUnit, exc: SootClass): Unit = {
    Util.debug(s"Adding try-catch around $u, for $exc")
    val excLoc = lg.generateLocal(exc.getType)
    val handlerStart: SootUnit =
      Jimple.v.newIdentityStmt(excLoc, Jimple.v.newCaughtExceptionRef)
    val handlerRecover: Seq[SootUnit] = u match {
      case assign: AssignStmt =>
        assign.getLeftOp match {
          case loc: Local if GlobalTransform.isInputType(loc.getType) =>
            val (dummyVal, stmts) = dummyValue(loc.getType)
            stmts ++ Seq(Jimple.v.newAssignStmt(loc, dummyVal))
          case _ =>
            Seq(Jimple.v.newNopStmt)
        }
      case _ => Seq(Jimple.v.newNopStmt)
    }
    val extraNop: SootUnit = Jimple.v.newNopStmt
    val followNop = Jimple.v.newNopStmt
    body.getUnits.insertAfter(Seq(extraNop, followNop).asJava, u)
    val goto: SootUnit = Jimple.v.newGotoStmt(followNop)
    body.getUnits.addLast(handlerStart)
    body.getUnits.insertAfter((handlerRecover ++ Seq(goto)).asJava, handlerStart)
    body.getTraps.add(Jimple.v.newTrap(exc, u, extraNop, handlerStart))
  }

  def isOutputMethod(m: SootMethod): Option[Level] = {
    val methName = m.toString
    val outputPatterns = Seq("println", "sendTextMessage")
     
    if (outputPatterns.exists (methName.contains(_))) {
      Some(L)
    } else {
      None
    }
  }

  def isInput(expr: InvokeExpr): Option[Level] =
    isInputMethod(expr.getMethod)

  def isOutput(expr: InvokeExpr): Option[Level] =
    isOutputMethod(expr.getMethod)

  def isApplicationMethod(m: SootMethod): Boolean = 
    Scene.v.getApplicationClasses.asScala.toList.contains(m.getClass())

  def newFacetedClassObj(t: Type): (Local, Seq[SootUnit]) = {
    val facetedClass = GlobalTransform.v.getFacetedClass(t)
    val facetedSootClass = facetedClass.getSootClass
    val fresultLoc = lg.generateLocal(facetedSootClass.getType)
    val fresultNewStmt = Jimple.v.newAssignStmt(fresultLoc,
      Jimple.v.newNewExpr(facetedSootClass.getType))
    val fresultInitStmt = Jimple.v.newInvokeStmt(
      Jimple.v.newSpecialInvokeExpr(fresultLoc,
        facetedClass.getConstructor.makeRef))
    (fresultLoc, Seq(fresultNewStmt, fresultInitStmt))
  }

  def createFacetedStruct(fv: FacetedVar): (Local, Seq[SootUnit]) = {
    val (floc, initStmts) = newFacetedClassObj(fv.getType)
    val facetedClass = GlobalTransform.v.getFacetedClass(fv.getType)
    val assignStmts = Lattice.levels map (l =>
      Jimple.v.newAssignStmt(
        Jimple.v.newInstanceFieldRef(floc, facetedClass.getLevelField(l)),
        fv.levelVal(l)))
    (floc, initStmts ++ assignStmts)
  }

  def dummyValue(t: Type): (Value, Seq[SootUnit]) = {
    def createWrapperObj(cls: SootClass,
      initVal: Value): (Value, Seq[SootUnit])= {
      val objLoc = lg.generateLocal(cls.getType)
      val newStmt = Jimple.v.newAssignStmt(objLoc, Jimple.v.newNewExpr(t.asInstanceOf[RefType]))
      val initMeth = cls.getMethod("<init>", Seq(initVal.getType).asJava)
      val initCall = Jimple.v.newInvokeStmt(Jimple.v.newSpecialInvokeExpr(objLoc,
        initMeth.makeRef, initVal))
      (objLoc, Seq(newStmt, initCall))
    }
    t match {
      case _ if t == IntType.v => (IntConstant.v(23), Seq())
      case _ if t == ShortType.v => (IntConstant.v(23), Seq())
      case _ if t == ByteType.v => (IntConstant.v(23), Seq())
      case _ if t == BooleanType.v => (IntConstant.v(0), Seq())
      case _ if t == LongType.v => (LongConstant.v(0), Seq())
      case _ if t == RefType.v("java.lang.String") =>
        (StringConstant.v("default value!!!"), Seq())
      case _ if t == DoubleType.v() => (DoubleConstant.v(23.0d), Seq())
      case _ if t == FloatType.v() => (FloatConstant.v(23.0f), Seq())
      case _ if t == RefType.v("java.lang.Float") =>
        createWrapperObj(Scene.v.loadClassAndSupport("java.lang.Float"),
          FloatConstant.v(23.0f))
      case _ if t == RefType.v("java.lang.Integer") =>
        createWrapperObj(Scene.v.loadClassAndSupport("java.lang.Integer"),
          IntConstant.v(23))
      case _ if t == CharType.v() => (IntConstant.v(64), Seq())
      case _ => Util.die(s"No dummy value for type $t")
    }
  }
}

object TaintTransformer {
  private val instance: TaintTransformer = new TaintTransformer()
  def v: TaintTransformer = return instance

  /* Generate a fresh copy of a variable such that all variable
   * names resulting from appending a level to a variable are
   * unused.
   */
  def freshVarCopy(v: String, existing: Set[String]): String = {
    val freshPref = v.toString()
    var cnt = -1
    var freshName = freshPref + "_"
    while (Lattice.levels.exists(l => existing.contains(freshName + l.toString()))) {
      Util.debug(s"Trying fresh name $freshName")
      cnt += 1
      freshName = freshPref + "_" + cnt
    }
    freshName
  }
}
